#!/usr/bin/env python
"""Dimensioned constants.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from numpy import array

from flightcondition.units import unit, dimless


class PhysicalConstants():
    """Dimensionalized constants for general physics.
    Air properties assume a calorically perfect gas in non-extreme conditions
    (approximately: T < 400 K and p < 20 psi).

    Attributes:
        g (acceleration): Acceleration due to gravity
        R (energy/temperature/mol): Universal gas constant
        R_air (energy/temperature/mass): Gas constant for air (non-extreme
            conditions)
        Cp_air (energy/temperature/mass): Specific heat at constant pressure
            for air (non-extreme conditions)
        Cv_air (energy/temperature/mass): Specific heat at constant volume
            for air (non-extreme conditions)
        gamma_air (dimless): Ratio of specific heats for air (low temperature
            and pressure conditions)
        collision_diam_air (length): Effective collision diameter of an air
            molecule
        N_A (1/mol): Avogadro's constant
        R_earth (length): Radius of Earth
    """
    g = 9.80665 * unit('m/s^2')
    R = 8.31432 * unit('N m / (mol K)')
    R_air = 287.05287 * unit('J/(K kg)')
    Cp_air = 1.005e3 * unit('J/(K kg)')
    Cv_air = 0.718e3 * unit('J/(K kg)')
    gamma_air = 1.4 * dimless
    Pr_air = 0.7 * dimless
    collision_diam_air = 0.365e-9 * unit('m')
    N_A = 6.02257e23 * unit('1/mol')
    R_earth = 6.356766e3 * unit('km')


class AtmosphereConstants():
    """Dimensionalized constants for atmospheric modeling.

    Attributes:
        p_0 (pressure): Sea level ambient pressure
        T_ice (temperature): Sea level ice point temperature
        T_0 (temperature): Sea level ambient temperature
        rho_0 (density): Sea level ambient density
        S (temperature): Sutherland's empirical constant in equation for
            dynamic viscosity
        beta_s (mass/length/time/temperature^0.5): Sutherland's empirical
            constant in equation for dynamic viscosity
    """

    # Sea level
    p_0 = 101.325e3 * unit('Pa')
    T_ice = 273.15 * unit('K')
    T_0 = 288.15 * unit('K')
    rho_0 = 1.225 * unit('kg/m^3')
    S = 110.4 * unit('K')
    beta_s = 1.458e-6 * unit('kg/(m*s*K^(1/2))')

    layer_names = array([
        "Troposphere",
        "Tropopause",
        "Stratosphere",
        "Stratosphere",
        "Stratopause",
        "Mesosphere",
        "Mesosphere",
        "Mesopause",
    ])

    # Layer base geopotential heights
    H_base = array([
        -5.0,
        0.0,
        11.0,
        20.0,
        32.0,
        47.0,
        51.0,
        71.0,
        80.0
    ]) * unit('km')

    # Layer temperature gradients
    T_grad = array([
        -6.5,
        -6.5,
        0.0,
        1.0,
        2.8,
        0.0,
        -2.8,
        -2.0,
        -2.0
    ]) * 1e-3 * unit('K/m')

    # Layer base temperatures
    T_base = array([
        320.65,
        288.15,
        216.65,
        216.65,
        228.65,
        270.65,
        270.65,
        214.65,
        196.65,
    ]) * unit('K')

    # Layer base pressures
    p_base = array([
        1.77687e+5,
        1.01325e+5,
        2.26320e+4,
        5.47487e+3,
        8.68014e+2,
        1.10906e+2,
        6.69384e+1,
        3.95639e+0,
        8.86272e-1,
    ]) * unit('Pa')



#!/usr/bin/env python
"""Provide dimensional units for mathematical operations using the pint
package.

Dependencies: pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from functools import wraps
from inspect import currentframe

from pint import UnitRegistry

__all__ = ['unit', 'dimless', 'printv']

unit = UnitRegistry(system='SI')
unit.default_format = '~P'
dimless = unit('dimensionless')

_US_length_units = ('ft', 'feet', 'foot', 'kft', 'kilofoot', 'kilofeet'
                    'fts', 'feets', 'foots', 'kfts', 'kilofoots', 'kilofeets',
                    'mi', 'mile', 'miles')


def check_same_registry(inp):
    """Check if input units are using this package's unit registry.

    Args:
        inp (object): Object to assert as dimensional type

    """
    if isinstance(inp, unit.Quantity):  # method to check if same registry
        dummy = 1 * unit('m')
        # Registry equality check
        return (inp._REGISTRY is dummy._REGISTRY)
    else:
        return False


def check_dimensioned(inp):
    """Check that input is dimensional (type 'Quantity' from pint package).

    Args:
        inp (object): Object to assert as dimensional type

    """
    try:
        inp.units.compatible_units()  # check that uses pint methods
    except AttributeError:
        raise TypeError("Input value is not correctly dimensioned! Use"
                        " dimensional type 'Quantity' from pint package.")


def check_dimensionless(inp):
    """Check that input is dimensionless type Quantity from pint package.

    Args:
        inp (object): Object to assert as length dimensional type

    """
    if not (inp.check('[]')):
        raise TypeError("Input value is not correctly typed! Use dimensionless"
                        " dimensional unit.")


def check_length_dimensioned(inp):
    """Check that input is length dimension type Quantity from pint package.

    Args:
        inp (object): Object to assert as length dimensional type

    """
    if not (inp.check('[length]')):
        raise TypeError("Input value is not correctly typed! Use length"
                        " dimensional unit.")


def check_US_length_units(ell):
    """Check if length unit type is an US unit

    Args:
        ell (length): Length unit of type Quantity from pint package.

    Returns:
        bool: True if US unit else False
    """
    return ell.units in _US_length_units


def check_area_dimensioned(inp):
    """Check that input is area type Quantity from pint package.

    Args:
        inp (object): Object to assert as area dimensional type

    """
    if not (inp.check('[length]^2')):
        raise TypeError("Input value is not correctly typed! Use area"
                        " dimensional unit.")


def name_of_var(var):
    """Find name of variables using local items.

    Args:
        var (unit): Dimensioned variable

    Returns:
        str: User-coded name of variable
    """
    try:
        local_vars = currentframe().f_back.f_back.f_locals.items()
        match = [name for name, val in local_vars if val is var]
    except AttributeError:
        local_vars = currentframe().f_back.f_locals.items()
        match = [name for name, val in local_vars if val is var]
    name = match[0] if match else "unknown"
    return name


def printv(var, to=None, name="", prec=".5g", fmt="~P", *args, **kwargs):
    """Print name and value of a Pint dimensional variable.
    For example,

        distance = 99.9 * unit('m')
        printv(distance)
        # prints "distance = 99.9 m"

    *Note*: As of Python 3.8, simply use the f-string syntax, e.g.
        x=7
        print(f"{x=}")

    Args:
        var (unit): Variable to be printed
        to (str): Convert to another unit
        name (str): Overwrite variable name
        prec (str): Precision formatter
        fmt (str): Additional formatter for pretty print or lack thereof
        *args: Additional arguments
        **kwargs: Additional keyword arguments

    """
    if isinstance(var, unit.Quantity):
        output = (var.to(to) if to is not None else var.to_base_units())
    else:
        output = var  # FIXME
    name = name if name else name_of_var(var)
    print(f"{name} = {output:{prec}{fmt}}", *args, **kwargs)


def to_base_units_wrapper(func):
    """Function decorator to convert output variable units to base units.

    Args:
        func (callable): Function to wrap

    Returns:
        callable: Called, wrapped function
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        if output is not None:
            output.ito_base_units()
        return output
    return wrapper



#!/usr/bin/env python
"""Compute atmospheric properties from ICAO 1993 standard atmosphere model.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

import warnings

import numpy as np

from flightcondition.constants import PhysicalConstants as Phys
from flightcondition.constants import AtmosphereConstants as Atmo
from flightcondition.common import AliasAttributes, DimensionalData,\
    _property_decorators
from flightcondition.units import unit, check_dimensioned,\
    check_length_dimensioned, check_US_length_units


class Layer(DimensionalData):
    """Class to compute and store layer data. """

    varnames = {
        'name': 'layer_name',
        'H_base': 'base_geopotential_height',
        'T_base': 'base_geopotential_temperature',
        'T_grad': 'temperature_gradient',
        'p_base': 'base_static_pressure',
    }

    def __init__(self, H_arr):
        """Initialize Layer nested class.

        Args:
            H_arr (length): Geopotential altitude

        """
        H_arr = np.atleast_1d(H_arr)

        self._layer_name = [""]*np.size(H_arr)
        self._H_base = np.zeros_like(H_arr) * unit('m')
        self._T_base = np.zeros_like(H_arr) * unit('K')
        self._T_grad = np.zeros_like(H_arr) * unit('K/m')
        self._p_base = np.zeros_like(H_arr) * unit('Pa')

        for idx, H in enumerate(H_arr):
            jdx = __class__._layer_idx(H)
            self._layer_name[idx] = Atmo.layer_names[jdx]
            self._H_base[idx] = Atmo.H_base[jdx]
            self._T_base[idx] = Atmo.T_base[jdx]
            self._T_grad[idx] = Atmo.T_grad[jdx]
            self._p_base[idx] = Atmo.p_base[jdx]

        # Initialize access by full quantity name through .byname.<name>
        self.byname = AliasAttributes(varsobj_arr=[self, ],
                                      varnames_dict_arr=[__class__.varnames, ])

    @staticmethod
    def _layer_idx(H):
        """Find index for layer data.

        Args:
            H_arr (length): Geopotential altitude

        Returns:
            int: Index of layer
        """
        idx = None
        for idx, H_base in enumerate(Atmo.H_base[:-1]):
            H_base_np1 = Atmo.H_base[idx+1]
            if H_base <= H < H_base_np1:
                return idx
        else:  # no break
            raise ValueError("H out of bounds.")

    def tostring(self, full_output=True):
        """Output string representation of class object.

        Args:
            full_output (bool): Set to True for full output

        Returns:
            str: String representation
        """
        return str(super().asdict())

    @property
    def name(self):
        """Layer name """
        if np.size(self._layer_name) == 1:
            layername = self._layer_name[0]
        else:
            layername = self._layer_name
        return layername

    @_property_decorators
    def H_base(self):
        """Layer base geopotential altitude :math:`H_{base}` """
        return self._H_base

    @_property_decorators
    def T_base(self):
        """Layer base temperature :math:`T_{base}` """
        return self._T_base

    @_property_decorators
    def T_grad(self):
        """Layer base temperature gradient :math:`T_{grad}` """
        return self._T_grad

    @_property_decorators
    def p_base(self):
        """Layer base pressure :math:`p_{base}` """
        return self._p_base


class Atmosphere(DimensionalData):
    """Compute quantities from International Civil Aviation Organization (ICAO)
    1993, which extends the US 1976 Standard Atmospheric Model to 80 km.

    Usage:
        from flightcondition import Atmosphere, unit

        # Compute atmospheric data for a scalar or array of altitudes
        h = [0.0, 44.2, 81.0] * unit('km')
        atm = Atmosphere(h)

        # Uncomment to print all atmospheric quantities:
        #print(f"\n{atm}")

        # Uncomment to print while specifying abbreviated output in US units:
        #print(f"\n{atm.tostring(full_output=False, units='US')}")

        # See also the linspace() function from numpy, e.g.
        # h = linspace(0, 81.0, 82) * unit('km')

        # Access individual properties and convert to desired units: "
        p, T, rho, nu, a, k = atm.p, atm.T, atm.rho, atm.nu, atm.a, atm.k
        print(f"\nThe pressure in psi is {p.to('psi'):.3g}")
        # >>> The pressure in psi is [14.7 0.024 0.000129] psi

        # Compute additional properties such as mean free path
        # Explore the class data structure for all options
        print( f"\nThe mean free path = {atm.MFP:.3g}")
        # >>> The mean free path = [7.25e-08 4.04e-05 0.00564] yd
    """

    varnames = {
        'h': 'geometric_altitude',
        'H': 'geopotential_altitude',
        'p': 'pressure',
        'T': 'temperature',
        'rho': 'density',
        'a': 'sound_speed',
        'mu': 'dynamic_viscosity',
        'nu': 'kinematic_viscosity',
        'k': 'thermal_conductivity',
        'g': 'gravity',
        'MFP': 'mean_free_path',
    }

    def __init__(self, h=None, units="", **kwargs):
        """Input geometric altitude - object contains the corresponding
        atmospheric quantities.

        Args:
            h (length): Geometric altitude - aliases are 'alt', 'altitude'
            units (str): Set to 'US' for US units or 'SI' for SI
        """
        # Compute altitude bounds
        self._H_min = Atmo.H_base[0]
        self._H_max = Atmo.H_base[-1]
        self._h_min = __class__._h_from_H(self._H_min)
        self._h_max = __class__._h_from_H(self._H_max)

        # Process altitude input
        # Check for hidden aliases
        h_aliases = ['alt', 'altitude']
        if h is None:
            h = __class__._arg_from_alias(h_aliases, kwargs)

        # Default to 0 kft
        if h is None:
            h = 0 * unit('ft')
        self.h = h

        # Process unit system
        if units not in dir(unit.sys):  # check if usable system
            if check_US_length_units(h):
                self.units = 'US'
            else:
                self.units = 'SI'
        else:
            self.units = units

        # Initialize access by full quantity name through .byname.<name>
        self.byname = AliasAttributes(varsobj_arr=[self, ],
                                      varnames_dict_arr=[__class__.varnames, ])

    def tostring(self, full_output=True, units=None, max_var_chars=0,
                 pretty_print=True):
        """String representation of data structure.

        Args:
            full_output (bool): Set to True for full output
            units (str): Set to 'US' for US units or 'SI' for SI
            max_var_chars (int): Maximum number of characters in unit string
            pretty_print (bool): Pretty print output

        Returns:
            str: String representation of class object
        """
        if units is not None:
            self.units = units

        if self.units == 'US':
            h_units   = 'kft'
            H_units   = 'kft'
            p_units   = 'lbf/ft^2'
            T_units   = 'degR'
            rho_units = 'slug/ft^3'
            a_units   = 'ft/s'
            mu_units  = 'lbf/ft^2 s'
            nu_units  = 'ft^2/s'
            k_units   = 'slug ft/s^3/degR'
            g_units   = 'ft/s^2'
            MFP_units = 'ft'
        else:  # SI units
            h_units   = 'km'
            H_units   = 'km'
            p_units   = 'Pa'
            T_units   = 'degK'
            rho_units = 'kg/m^3'
            a_units   = 'm/s'
            mu_units  = 'Pa s'
            nu_units  = 'm^2/s'
            k_units   = 'W/m/K'
            g_units   = 'm/s^2'
            MFP_units = 'm'

        # Insert longer variable name into output
        max_var_chars = max([
            max([len(v) for v in __class__.varnames.values()]),
            max_var_chars
        ])
        h_str = self._vartostr(var=self.h, var_str='h', to_units=h_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.5g", pretty_print=pretty_print)
        H_str = self._vartostr(var=self.H, var_str='H', to_units=H_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.5g", pretty_print=pretty_print)
        p_str = self._vartostr(var=self.p, var_str='p', to_units=p_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.5g", pretty_print=pretty_print)
        T_str = self._vartostr(var=self.T, var_str='T', to_units=T_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.5g", pretty_print=pretty_print)
        rho_str = self._vartostr(var=self.rho, var_str='rho',
                                 to_units=rho_units,
                                 max_var_chars=max_var_chars,
                                 fmt_val="10.5g", pretty_print=pretty_print)
        a_str = self._vartostr(var=self.a, var_str='a', to_units=a_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.5g", pretty_print=pretty_print)
        mu_str = self._vartostr(var=self.mu, var_str='mu', to_units=mu_units,
                                max_var_chars=max_var_chars,
                                fmt_val="10.4e", pretty_print=pretty_print)
        nu_str = self._vartostr(var=self.nu, var_str='nu', to_units=nu_units,
                                max_var_chars=max_var_chars,
                                fmt_val="10.4e", pretty_print=pretty_print)
        k_str = self._vartostr(var=self.k, var_str='k', to_units=k_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.4e", pretty_print=pretty_print)
        g_str = self._vartostr(var=self.g, var_str='g', to_units=g_units,
                               max_var_chars=max_var_chars,
                               fmt_val="10.5g", pretty_print=pretty_print)
        MFP_str = self._vartostr(var=self.MFP, var_str='MFP',
                                 to_units=MFP_units,
                                 max_var_chars=max_var_chars,
                                 fmt_val="10.4e", pretty_print=pretty_print)

        if full_output:
            if type(self.layer.name) is np.str_:  # singular string
                trunc_layer_name = self.layer.name
            else:
                trunc_layer_name = "[" + " ".join([
                    f"{s[:10]}" for s in self.layer.name
                ]) + "]"
            layer_str = (f"{'atmospheric_layer ':{max_var_chars}} name    = "
                         f"{trunc_layer_name}")

            repr_str = (f"{h_str}\n{H_str}\n{p_str}\n{T_str}\n{rho_str}\n"
                        f"{a_str}\n{mu_str}\n{nu_str}\n{k_str}\n{g_str}\n"
                        f"{MFP_str}\n{layer_str}")
        else:
            repr_str = (f"{h_str}\n{p_str}\n{T_str}\n{rho_str}\n{a_str}\n"
                        f"{nu_str}")
        return repr_str

    @staticmethod
    @unit.wraps(unit.m, (unit.m, unit.m))
    def _H_from_h(h, R_earth=Phys.R_earth):
        """Convert geometric to geopotential altitude.

        :math:`H = \\frac{R_{earth} h}{R_{earth} + h}`

        Args:
            h (length): Geometric altitude
            R_earth (length): Radius of the Earth

        Returns:
            length: Geopotential altitude
        """
        H = R_earth*h/(R_earth + h)
        return H

    @staticmethod
    @unit.wraps(unit.m, (unit.m, unit.m))
    def _h_from_H(H, R_earth=Phys.R_earth):
        """Convert geopotential to geometric altitude.

        :math:`h = \\frac{R_{earth} H}{R_{earth} - H}`

        Args:
            H (length): Geopotential altitude
            R_earth (length): Radius of the Earth

        Returns:
            length: Geometric altitude
        """
        h = R_earth*H/(R_earth - H)
        return h

    @staticmethod
    @unit.wraps(unit('W/m/K').units, unit.K)
    def _thermal_conductivity(T_K):
        """Compute thermal conductivity.

        Args:
            T_K (temperature): Dimensional Temperature at altitude, which is
                automatically converted to Kelvin

        Returns:
            power/length/temperature: Geometric altitude in W/m/K
        """
        # Empirical formula requires T in Kelvin
        c1 = 0.002648151
        k = c1*T_K**1.5 / (T_K + (245.4*10**(-12/T_K)))
        return k

    @property
    def units(self):
        """Get unit system to use: 'SI', 'US', etc.  Available unit systems
        given by dir(unit.sys).

        Returns:
            str: Unit system
        """
        return self._units

    @units.setter
    def units(self, units):
        """Set unit system to use: 'SI', 'US', etc.  Available unit systems
        given by dir(unit.sys).

        Args:
            units (str): Unit system
        """
        if units not in dir(unit.sys):
            warnings.warn(f"'{units} is not available. Try one of the "
                          f"following: {dir(unit.sys)}")
            return
        else:
            self._units = units
            unit.default_system = units

    @_property_decorators
    def h(self):
        """Get geometric altitude :math:`h`

        Returns:
            length: Geometric altitude
        """
        return self._h

    @h.setter
    def h(self, h):
        """Set geometric altitude :math:`h`

        Check that input is of type Quantity from pint package. Check that
        input is length dimension.  Check bounds.  Format as array even if
        scalar input.

        Args:
            h (length): Input scalar or array of altitudes
        """
        tofloat = 1.0
        h = np.atleast_1d(h) * tofloat

        check_dimensioned(h)
        h = h.magnitude * unit(str(h.units))  # force local unit registry
        check_length_dimensioned(h)

        if len(np.shape(h)) > 1:
            raise TypeError("Input must be scalar or 1-D array.")

        if (h < self._h_min).any() or (self._h_max < h).any():
            raise ValueError(
                f"Input altitude is out of bounds "
                f"({self._h_min:.5g} < h < {self._h_max:.5g})"
            )

        # Update quantities
        self._h = h
        self._H = self._H_from_h(self.h)
        self.layer = Layer(self.H)

    @_property_decorators
    def H(self):
        """Geopotential altitude :math:`H` """
        return self._H

    @_property_decorators
    def p(self):
        """Air pressure :math:`p` """
        H_base = np.atleast_1d(self.layer.H_base)
        T_base = np.atleast_1d(self.layer.T_base)
        T_grad = np.atleast_1d(self.layer.T_grad)
        p_base = np.atleast_1d(self.layer.p_base)

        H = np.atleast_1d(self.H)
        T = np.atleast_1d(self.T)
        g_0 = Phys.g
        R_air = Phys.R_air

        p = np.zeros_like(H) * unit('Pa')

        # Pressure equation changes between T_grad == 0 and T_grad != 0
        s = T_grad == 0
        p[s] = p_base[s]*np.exp((-g_0/(R_air*T[s]))*(H[s] - H_base[s]))

        s = T_grad != 0
        p[s] = p_base[s]*(
            1 + (T_grad[s]/T_base[s])*(H[s] - H_base[s])
        )**((1/T_grad[s])*(-g_0/R_air))

        return p

    @_property_decorators
    def T(self):
        """Ambient air temperature :math:`T` """
        T_grad = np.atleast_1d(self.layer.T_grad)
        H_base = np.atleast_1d(self.layer.H_base)
        T_base = np.atleast_1d(self.layer.T_base)
        T = T_base + T_grad*(self.H - H_base)
        return T

    @_property_decorators
    def rho(self):
        """Ambient air density :math:`\\rho` """
        p = self.p
        T = self.T
        R_air = Phys.R_air
        rho = p/(R_air*T)  # TODO 2022-08-07: separate out
        return rho

    @_property_decorators
    def a(self):
        """Ambient speed of sound :math:`a` """
        T = self.T
        gamma_air = Phys.gamma_air
        R_air = Phys.R_air
        a = np.sqrt(gamma_air*R_air*T)  # TODO 2022-08-07: separate out
        return a

    @_property_decorators
    def mu(self):
        """Ambient dynamic viscosity :math:`\\mu` """
        T = self.T
        beta_s = Atmo.beta_s
        S = Atmo.S
        mu = beta_s*T**1.5/(T + S)
        return mu

    @_property_decorators
    def nu(self):
        """Ambient kinematic viscosity :math:`\\nu` """
        mu = self.mu
        rho = self.rho
        nu = mu/rho
        return nu

    @_property_decorators
    def k(self):
        """Ambient thermal conductivity :math:`k` """
        k = __class__._thermal_conductivity(self.T)
        return k

    @_property_decorators
    def g(self):
        """Gravitational acceleration at altitude :math:`g` """
        h = self.h
        g_0 = Phys.g
        R_earth = Phys.R_earth
        g = g_0*(R_earth/(R_earth + h))**2
        return g

    @_property_decorators
    def MFP(self):
        """Mean free path """
        p = self.p
        T = self.T
        N_A = Phys.N_A
        R = Phys.R
        sigma = Phys.collision_diam_air

        n = N_A*p/(R*T)  # number density
        MFP = 1/(np.sqrt(2)*np.pi*sigma**2*n)
        return MFP



#!/usr/bin/env python
"""Easily convert between Mach number, true airspeed (TAS), calibrated airspeed
(CAS), and equivalent airspeed (EAS) for given altitude(s).  Additional flight
condition data and atmospheric data is computed.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

import warnings

import numpy as np

from flightcondition.atmosphere import Atmosphere
from flightcondition.boundarylayer import BoundaryLayer
from flightcondition.constants import PhysicalConstants as Phys
from flightcondition.common import AliasAttributes, DimensionalData,\
    _len1array_to_scalar, _property_decorators
from flightcondition.isentropicflow import IsentropicFlow
from flightcondition.nondimensional import NonDimensional
from flightcondition.units import unit, dimless, check_area_dimensioned,\
    check_dimensioned, check_dimensionless, check_length_dimensioned,\
    check_US_length_units, to_base_units_wrapper


class Velocity(DimensionalData):
    """Class to hold airspeed data. """

    varnames = {
        'TAS': 'true_airspeed',
        'CAS': 'calibrated_airspeed',
        'EAS': 'equivalent_airspeed',
        'M': 'mach_number',
        'mu_M': 'mach_angle',
        'q_inf': 'dynamic_pressure',
        'q_c': 'impact_pressure',
        'p0': 'stagnation_pressure',
        'T0': 'stagnation_temperature',
        'Tr_lamr': 'recovery_temperature_laminar',
        'Tr_turb': 'recovery_temperature_turbulent',
        'Re_by_L': 'reynolds_per_length',
    }

    def __init__(self, byalt):
        """Initialize.

        Args:
            byalt (object): Altitude object
        """
        # Link to Atmosphere data
        self._byalt = byalt
        h0 = 0 * unit('kft')
        self._atm0 = Atmosphere(h0)

    def tostring(self, full_output=True, units=None, max_var_chars=0,
                 pretty_print=True):
        """String representation of data structure.

        Args:
            full_output (bool): Set to True for full output
            units (str): Set to 'US' for US units or 'SI' for SI
            max_var_chars (int): Maximum number of characters in unit string
            pretty_print (bool): Pretty print output

        Returns:
            str: String representation
        """
        # Set default unit system
        if units is None:
            units = self._byalt.units

        if units == 'US':
            EAS_units   = 'knots'
            TAS_units   = 'knots'
            CAS_units   = 'knots'
            q_inf_units     = 'lbf/ft^2'
            q_c_units     = 'lbf/ft^2'
            p0_units    = 'lbf/ft^2'
            T0_units    = 'degR'
            Tr_lamr_units = 'degR'
            Tr_turb_units = 'degR'
            Re_by_L_units = '1/in'
        else:  # default to SI units
            TAS_units   = 'm/s'
            CAS_units   = 'm/s'
            EAS_units   = 'm/s'
            q_inf_units     = 'Pa'
            q_c_units     = 'Pa'
            p0_units    = 'Pa'
            T0_units    = 'degK'
            Tr_lamr_units = 'degK'
            Tr_turb_units = 'degK'
            Re_by_L_units = '1/mm'

        # Insert longer variable name into output
        max_var_chars = max([
            max([len(v) for v in self.varnames.values()]),
            max_var_chars
        ])
        M_str = self._vartostr(
            var=self.M, var_str='M', to_units='',
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        TAS_str = self._vartostr(
            var=self.TAS, var_str='TAS', to_units=TAS_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        CAS_str = self._vartostr(
            var=self.CAS, var_str='CAS', to_units=CAS_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        EAS_str = self._vartostr(
            var=self.EAS, var_str='EAS', to_units=EAS_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        q_inf_str = self._vartostr(
            var=self.q_inf, var_str='q_inf', to_units=q_inf_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        q_c_str = self._vartostr(
            var=self.q_c, var_str='q_c', to_units=q_c_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        p0_str = self._vartostr(
            var=self.p0, var_str='p0', to_units=p0_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        T0_str = self._vartostr(
            var=self.T0, var_str='T0', to_units=T0_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        Tr_lamr_str = self._vartostr(
            var=self.Tr_lamr, var_str='Tr_lamr', to_units=Tr_lamr_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        Tr_turb_str = self._vartostr(
            var=self.Tr_turb, var_str='Tr_turb', to_units=Tr_turb_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print) + "\n"
        Re_by_L_str = self._vartostr(
            var=self.Re_by_L, var_str='Re_by_L', to_units=Re_by_L_units,
            max_var_chars=max_var_chars, fmt_val="10.4e",
            pretty_print=pretty_print)

        if np.isnan(np.atleast_1d(self.mu_M)).all():
            mu_M_str = ""
        else:
            mu_M_str = self._vartostr(
                var=self.mu_M, var_str='mu_M', to_units='deg',
                max_var_chars=max_var_chars, fmt_val="10.5g",
                pretty_print=pretty_print) + "\n"

        if full_output:
            repr_str = (f"{TAS_str}{CAS_str}{EAS_str}{M_str}{mu_M_str}"
                        f"{q_inf_str}{q_c_str}{p0_str}{T0_str}{Tr_lamr_str}"
                        f"{Tr_turb_str}{Re_by_L_str}")
        else:
            repr_str = (f"{TAS_str}{CAS_str}{EAS_str}{M_str}{Re_by_L_str}")

        return repr_str

    @staticmethod
    def _compare_input_size(input_a, input_b):
        """Enforce non-singular input arrays to be equal in size

        Args:
            input_a (object): Input A
            input_b (object): Input B

        Returns:
            object: sized_arr
        """
        wrongsize_msg = ("Non-singular input arrays must be equal in size.")

        if np.shape(input_b):  # if array
            if np.shape(input_a):  # if array
                if not input_a.size == 1 and not input_b.size == 1:
                    if not input_a.size == input_b.size:
                        raise TypeError(wrongsize_msg)
            sizedarr = np.ones(np.shape(input_b))*input_a
        else:
            sizedarr = input_a
        return sizedarr

    def _check_and_size_input(self, input_var, input_alt=None, input_vel=None):
        """Check that input is correctly typed, then size input array. If
        scalar, leave as scalar.

        Args:
            input_var (object): Input array (or scalar)
            input_alt (object): Input array of altitude level
            input_vel (object): Input array of velocity level (when assessing
                length scale as input_var)

        Returns:
            object: Sized array (or scalar)
        """
        check_dimensioned(input_var)

        # Force local unit registry
        input_var = input_var.magnitude * unit(str(input_var.units))

        # Require input, base, and secondary arrays to be singular or
        # non-singular but equal in size
        sizedarr = __class__._compare_input_size(input_var, input_alt)
        if input_vel is not None:
            sizedarr = __class__._compare_input_size(input_var, input_vel)

        tofloat = 1.0
        sizedarr = (sizedarr * tofloat)

        return sizedarr

    @staticmethod
    def _impact_pressure(M, p):
        """Equation for impact pressure

        Args:
            M (dimless): Mach number
            p (pressure): Static pressure

        Returns:
            pressure: Impact pressure
        """
        # p0 = p + q_c
        y = Phys.gamma_air
        p0 = IsentropicFlow.p0_by_p(M, y)*p
        q_c = p0 - p
        return q_c

    @to_base_units_wrapper
    def _M_from_TAS(self, TAS):
        """Compute Mach number from true airspeed.

        Args:
            TAS (speed): True airspeed

        Returns:
            dimless: Mach number
        """
        a_inf = self._byalt.a
        M = NonDimensional.mach_number(U=TAS, a=a_inf)
        return M

    @to_base_units_wrapper
    def _TAS_from_M(self, M):
        """Compute true airspeed from Mach number.

        Args:
            M (dimless): Mach number

        Returns:
            speed: True airspeed
        """
        a_inf = self._byalt.a
        TAS = NonDimensional.mach_velocity(M, a_inf)
        return TAS

    @to_base_units_wrapper
    def _M_from_EAS(self, EAS):
        """Computer Mach number from equivalent airspeed.

        Args:
            EAS (speed): Equivalent airspeed

        Returns:
            dimless: Mach number
        """
        a_inf_h0 = self._atm0.a
        p_inf_h0 = self._atm0.p
        p_inf = self._byalt.p
        delta = p_inf/p_inf_h0
        M = EAS/(a_inf_h0*np.sqrt(delta))
        return M

    @to_base_units_wrapper
    def _EAS_from_M(self, M):
        """Computer Mach number from equivalent airspeed.

        Args:
            M (dimless): Mach number

        Returns:
            speed: Equivalent airspeed
        """
        a_inf_h0 = self._atm0.a
        p_inf_h0 = self._atm0.p
        p_inf = self._byalt.p
        delta = p_inf/p_inf_h0
        EAS = M*(a_inf_h0*np.sqrt(delta))
        return EAS

    @to_base_units_wrapper
    def _q_c_from_CAS(self, CAS):
        """Compute impact pressure from calibrated airspeed (accounting for
           compressibility).

        Args:
            CAS (speed): Calibrated airspeed

        Returns:
            pressure: Impact pressure
        """
        a_h0 = self._atm0.a
        p_h0 = self._atm0.p
        # Account for compressibility with the isentropic flow equation
        M_ = NonDimensional.mach_number(U=CAS, a=a_h0)
        q_c = __class__._impact_pressure(M=M_, p=p_h0)
        return q_c

    @to_base_units_wrapper
    def _CAS_from_q_c(self, q_c):
        """Compute calibrated airspeed from impact pressure (accounting for
           compressibility).

        Args:
            CAS (speed): Calibrated airspeed

        Returns:
            pressure: Impact pressure

        """
        a_h0 = self._atm0.a
        p_h0 = self._atm0.p
        # Account for compressibility with the isentropic flow equation
        # M_should = __class__._isentropic_mach(p0=q_c, p=p_h0)  # DEPRECATED
        M = IsentropicFlow.M_from_p0_by_p((q_c+p_h0)/p_h0)
        CAS = NonDimensional.mach_velocity(M, a_h0)  # subsonic
        return CAS

    @to_base_units_wrapper
    def _M_from_q_c(self, q_c):
        """Compute Mach number from impact pressure.

        Args:
            q_c (pressure): Impact pressure

        Returns:
            dimless: Mach number
        """
        p_inf = self._byalt.p
        # Isentropic flow equation
        # M_should = __class__._isentropic_mach(p0=q_c, p=p_inf)  # DEPRECATED
        M = IsentropicFlow.M_from_p0_by_p((q_c+p_inf)/p_inf)

        return M

    @to_base_units_wrapper
    def _q_c_from_M(self, M):
        """Compute impact pressure from Mach number.

        Args:
            M: Mach number

        Returns:
            impact pressure
        """
        p_inf = self._byalt.p
        # Solve for impact pressure from isentropic flow equation:
        q_c = __class__._impact_pressure(M=M, p=p_inf)
        return q_c

    @to_base_units_wrapper
    def _EAS_from_TAS(self, TAS, M):
        """Convert airspeed to true calibrated airspeed.

        Args:
            TAS (speed): True airspeed
            M (dimless): Mach number

        Returns:
            speed: Equivalent airspeed
        """
        a_inf_h0 = self._atm0.a
        p_inf_h0 = self._atm0.p
        p_inf = self._byalt.p
        delta = p_inf/p_inf_h0
        EAS = M*(a_inf_h0*np.sqrt(delta))
        return EAS

    @staticmethod
    @_len1array_to_scalar
    @unit.wraps(unit.rad, dimless.units)
    def _mach_angle(M):
        """Compute Mach angle

        Args:
            M (dimless): Mach number

        Returns:
            angle: mach angle
        """
        # Create new array and split computation to subsonic and supersonic
        M = np.atleast_1d(M)
        mu_M = M*0.0

        sub = M < 1  # subsonic filter
        if sub.any():
            mu_M[sub] = np.nan

        sup = M >= 1  # supersonic filter
        mu_M[sup] = np.arcsin(1.0/M[sup])

        return mu_M

    @staticmethod
    @to_base_units_wrapper
    def _q_inf_from_TAS(TAS, rho):
        """Compute dynamic pressure from true airspeed.

        Args:
            TAS (speed): True airspeed
            rho (density): Ambient density

        Returns:
            pressure: Dynamic pressure
        """
        q_inf = 0.5*rho*TAS**2
        return q_inf

    @_property_decorators
    def M(self):
        """Get Mach number :math:`M` """
        return self._M

    @M.setter
    def M(self, M):
        """Mach number :math:`M` """
        M *= dimless  # add dimless for raw float input
        self._M = self._check_and_size_input(M, input_alt=self._byalt.h)
        self._TAS = self._TAS_from_M(self.M)
        self._EAS = self._EAS_from_TAS(self.TAS, self.M)
        self._q_c = self._q_c_from_M(self.M)
        self._CAS = self._CAS_from_q_c(self.q_c)

    @_property_decorators
    def TAS(self):
        """Get true airspeed. """
        return self._TAS

    @TAS.setter
    def TAS(self, TAS):
        """Set true airspeed. """
        self._TAS = self._check_and_size_input(TAS, input_alt=self._byalt.h)
        self._M = self._M_from_TAS(TAS)
        self._EAS = self._EAS_from_TAS(self.TAS, self.M)
        self._q_c = self._q_c_from_M(self.M)
        self._CAS = self._CAS_from_q_c(self.q_c)

    @_property_decorators
    def CAS(self):
        """Get calibrated airspeed. """
        return self._CAS

    @CAS.setter
    def CAS(self, CAS):
        """Calibrated airspeed. """
        self._CAS = self._check_and_size_input(CAS, input_alt=self._byalt.h)
        self._q_c = self._q_c_from_CAS(self.CAS)
        self._M = self._M_from_q_c(self.q_c)
        self._TAS = self._TAS_from_M(self.M)
        self._EAS = self._EAS_from_TAS(self.TAS, self.M)

    @_property_decorators
    def EAS(self):
        """Get equivalent airspeed. """
        return self._EAS

    @EAS.setter
    def EAS(self, EAS):
        """Set equivalent airspeed. """
        self._EAS = self._check_and_size_input(EAS, input_alt=self._byalt.h)
        self._M = self._M_from_EAS(self.EAS)
        self._TAS = self._TAS_from_M(self.M)
        self._q_c = self._q_c_from_M(self.M)
        self._CAS = self._CAS_from_q_c(self.q_c)

    @_property_decorators
    def mu_M(self):
        """Get Mach angle :math:`\\mu_M` """
        return __class__._mach_angle(self.M)

    @_property_decorators
    def q_c(self):
        """Get impact pressure :math:`q_c`"""
        return self._q_c

    @_property_decorators
    def q_inf(self):
        """Get dynamic pressure :math:`q_\\infty`"""
        q_inf = __class__._q_inf_from_TAS(TAS=self.TAS, rho=self._byalt.rho)
        return q_inf

    @_property_decorators
    def p0(self):
        """Get stagnation pressure :math:`p_0`"""
        M = self.M
        p = self._byalt.p
        y = Phys.gamma_air
        p0 = IsentropicFlow.p0_by_p(M, y)*p
        return p0

    @_property_decorators
    def T0(self):
        """Get stagnation temperature :math:`T_0`"""
        M = self.M
        T = self._byalt.T
        y = Phys.gamma_air
        T0 = IsentropicFlow.T0_by_T(M, y)*T
        return T0

    @_property_decorators
    def Tr_lamr(self):
        """Get recovery temperature (laminar) :math:`Tr_{laminar}`"""
        Tr_lamr = BoundaryLayer.recovery_temperature_laminar(
            M=self.M, T=self._byalt.T)
        return Tr_lamr

    @_property_decorators
    def Tr_turb(self):
        """Get recovery temperature (turbulent) :math:`Tr_{turbulent}`"""
        Tr_turb = BoundaryLayer.recovery_temperature_turbulent(
            M=self.M, T=self._byalt.T)
        return Tr_turb

    @_property_decorators
    def Re_by_L(self):
        """Get Reynolds number per unit length :math:`Re_L`"""
        length_unit = 'in' if self._byalt.units == 'US' else None
        Re_by_L = NonDimensional.reynolds_per_length(U=self.TAS,
                                                     nu=self._byalt.nu,
                                                     length_unit=length_unit)
        return Re_by_L


class Length(DimensionalData):
    """Class to hold length data. """

    varnames = {
        'L': 'length_scale',
        'Re': 'reynolds_number',
        # 'Kn': 'knudsen_number',
        'h_BL_turb': 'boundary_thickness_turbulent',
        'h_BL_lamr': 'boundary_thickness_laminar',
        'Cf_lamr': 'friction_coefficient_laminar',
        'Cf_turb': 'friction_coefficient_turbulent',
        'h_yplus1': 'wall_distance_yplus1',
    }

    def __init__(self, byalt, byvel):
        """Initialize.

        Args:
            byalt (object): Altitude object
            byvel (object): Velocity object
        """
        # Link to Atmosphere and Velocity data
        self._byalt = byalt
        self._byvel = byvel

    def tostring(self, full_output=True, units=None, max_var_chars=0,
                 pretty_print=True):
        """String representation of data structure.

        Args:
            full_output (bool): Set to True for full output
            units (str): Set to 'US' for US units or 'SI' for SI
            max_var_chars (int): Maximum number of characters in unit string
            pretty_print (bool): Pretty print output

        Returns:
            str: String representation
        """
        # Set default unit system
        if units is None:
            units = self._byalt.units

        if units == 'US':
            L_units = 'ft'
            h_BL_units = 'in'
        else:  # default to SI units
            L_units = 'm'
            h_BL_units = 'mm'

        # Insert longer variable name into output
        max_var_chars = max([
            max([len(v) for v in self.varnames.values()]),
            max_var_chars
        ])
        L_str = self._vartostr(
            var=self.L, var_str='L', to_units=L_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)
        Re_str = self._vartostr(
            var=self.Re, var_str='Re', to_units='',
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)
        h_BL_lamr_str = self._vartostr(
            var=self.h_BL_lamr, var_str='h_BL_lamr', to_units=h_BL_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)
        h_BL_turb_str = self._vartostr(
            var=self.h_BL_turb, var_str='h_BL_turb', to_units=h_BL_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)
        Cf_lamr_str = self._vartostr(
            var=self.Cf_lamr, var_str='Cf_lamr', to_units=dimless,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)
        Cf_turb_str = self._vartostr(
            var=self.Cf_turb, var_str='Cf_turb', to_units=dimless,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)
        h_yplus1_str = self._vartostr(
            var=self.h_yplus1, var_str='h_yplus1', to_units=h_BL_units,
            max_var_chars=max_var_chars, fmt_val="10.5g",
            pretty_print=pretty_print)

        if full_output:
            repr_str = (f"{L_str}\n{Re_str}\n{h_BL_lamr_str}\n"
                        f"{h_BL_turb_str}\n{Cf_lamr_str}\n{Cf_turb_str}\n"
                        f"{h_yplus1_str}")
        else:
            repr_str = (f"{L_str}\n{Re_str}")

        return repr_str

    @to_base_units_wrapper
    def wall_distance_from_yplus(self, yplus):
        """Compute wall distance from yplus value.

        :math:`y^+ = \\frac{y u_\\tau}{\\nu}`

        where

        :math:`u_\\tau = \\sqrt{\\frac{\\tau_w}{\\rho}}

        Args:
            yplus (dimless): yplus value

        Returns:
            length: Distance from wall
        """
        yplus *= dimless
        check_dimensionless(yplus)

        rho = self._byalt.rho
        nu = self._byalt.nu
        q_inf = self._byvel.q_inf
        with warnings.catch_warnings():  # catch warning for divide by 0
            warnings.simplefilter("ignore")
            Cf_turb = self.Cf_turb

            tau_wall = Cf_turb*q_inf
            u_tau = np.sqrt(tau_wall/rho)
            y = yplus*nu/u_tau

        return y

    @_property_decorators
    def L(self):
        """Get length scale :math:`L`"""
        return self._L

    @L.setter
    def L(self, L):
        """Set length scale :math:`L`"""
        # Verify that length input is dimensional length quantity
        L = self._byvel._check_and_size_input(L, input_alt=self._byalt.h,
                                              input_vel=self._byvel.M)
        check_length_dimensioned(L)
        self._L = L

    @_property_decorators
    def Re(self):
        """Get Reynolds number :math:`Re`"""
        Re = NonDimensional.reynolds_number(
            U=self._byvel.TAS, L=self.L, nu=self._byalt.nu)
        return Re

    @Re.setter
    def Re(self, Re):
        """Set Reynolds number :math:`Re`
        Set the true airspeed based on Reynolds number and length scale. """
        Re *= dimless  # add dimless for raw float input
        Re = self._byvel._check_and_size_input(
            Re, input_alt=self._byalt.h, input_vel=self._byvel.M)
        self._byvel.TAS = NonDimensional.reynolds_number_velocity(
            Re_L=Re, L=self.L, nu=self._byalt.nu)

    # @_property_decorators  # disable for now
    # def Kn(self):
    #     """Get Knudsen number :math:`K_n`"""
    #     Kn = NonDimensional.knudsen_number(self.L, self._byalt.MFP)
    #     return Kn

    @_property_decorators
    def h_BL_lamr(self):
        """Get laminar Boundary Layer Thickness :math:`\\h_BL_{lamr}` """
        x = self.L
        Re_x = NonDimensional.reynolds_number(
            U=self._byvel.TAS, L=x, nu=self._byalt.nu)
        h_BL_lamr = BoundaryLayer.flat_plate_boundary_layer_lamr(
            x=self.L, Re_x=Re_x)
        return h_BL_lamr

    @_property_decorators
    def h_BL_turb(self):
        """Get turbulent Boundary Layer Thickness :math:`\\h_BL_{turb}` """
        x = self.L
        Re_x = NonDimensional.reynolds_number(
            U=self._byvel.TAS, L=x, nu=self._byalt.nu)
        h_BL_turb = BoundaryLayer.flat_plate_boundary_layer_turb(
            x=self.L, Re_x=Re_x)
        return h_BL_turb

    @_property_decorators
    def Cf_lamr(self):
        """Get laminar skin friction coefficient :math:`\\Cf_{lamr}` """
        x = self.L
        Re_x = NonDimensional.reynolds_number(
            U=self._byvel.TAS, L=x, nu=self._byalt.nu)
        Cf_lamr = BoundaryLayer.flat_plate_skin_friction_coeff_lamr(Re_x=Re_x)
        return Cf_lamr

    @_property_decorators
    def Cf_turb(self):
        """Get turbulent skin friction coefficient :math:`\\Cf_{turb}` """
        x = self.L
        Re_x = NonDimensional.reynolds_number(
            U=self._byvel.TAS, L=x, nu=self._byalt.nu)
        Cf_turb = BoundaryLayer.flat_plate_skin_friction_coeff_turb(Re_x=Re_x)
        return Cf_turb

    @_property_decorators
    def h_yplus1(self):
        """Get height from wall where yplus=1. """
        h_yplus1 = self.wall_distance_from_yplus(1)
        return h_yplus1


class FlightCondition(DimensionalData):
    """Easily convert between Mach number, true airspeed (TAS), calibrated
    airspeed (CAS), and equivalent airspeed (EAS) for given altitude(s).
    Additional flight condition data and atmospheric data is computed.

    All inputs must be dimensional unit quantities.

    Usage:
        from flightcondition import FlightCondition, unit, dimless

        # Compute flight condition at 3 km, Mach 0.5
        fc = FlightCondition(3*unit('km'), M=0.5)

        # Uncomment to print summary of flight condition quantities:
        #print(f"{fc}")

        # Uncomment to print abbreviated output in US units:
        #print(f"\n{fc.tostring(full_output=False, units='US')}")

        # Access true, calibrated, equivalent airspeeds
        KTAS = fc.byvel.TAS.to('knots')
        KCAS = fc.byvel.CAS.to('knots')
        KEAS = fc.byvel.EAS.to('knots')
        print(f"Flying at {KTAS.magnitude:.4g} KTAS,"
            f" which is {KCAS.magnitude:.4g} KCAS,"
            f" or {KEAS.magnitude:.4g} KEAS")
        # >>> Flying at 319.4 KTAS, which is 277.7 KCAS, or 275.1 KEAS

        # Access atmospheric data (see Atmosphere class for more)
        atm = fc.byalt  # access Atmosphere object
        h, p, T, rho, nu, a = atm.h, atm.p, atm.T, atm.rho, atm.nu, atm.a
        print(f"The ambient temperature at {h.to('km'):.4g} is {T:.4g}")
        # >>> The ambient temperature at 3 km is 268.7 K

        # Compute again instead using true airspeed and altitude in km
        fc = FlightCondition(3.048*unit('km'), TAS=401.7*unit('mph'))
        #print(f"{fc}")  # uncomment to print output

        # Compute for a range of altitudes at 275.14 knots-equivalent
        # airspeed with a characteristic length scale of 10 meters
        fc = FlightCondition([0, 9.8425, 20]*unit('kft'),
                            EAS=275.14*unit('kt'),
                            L=10*unit('m'))

        # Compute additional derived quantities
        # Explore the class data structure for all options
        print(f"\nThe dynamic pressure in psi is "
            f"{fc.byvel.q_inf.to('psi'):.3g}")
        # >>> The dynamic pressure in psi is [1.78 1.78 1.78] psi
        print(f"The Reynolds number is {fc.bylen.Re:.3g}")
        # >>> The Reynolds number is [9.69e+07 8.82e+07 7.95e+07]

        # Alternatively access quantities by their full name
        print(fc.byvel.TAS == fc.byname.true_airspeed)
        # >>> [ True  True  True]
    """

    def __init__(
        self, h=None, M=None, TAS=None, CAS=None, EAS=None, L=None, Re=None,
        units="", **kwargs,
    ):
        """Constructor based on altitude and input velocity in terms of Mach
        number, TAS, CAS, or EAS.  Input altitude, one format of velocity, and
        length scale.  Reynolds number can be input as an alternative to either
        velocity or length scale but not both.  All inputs must be dimensional
        unit quantities.

        Args:
            h (length): Geometric altitude - aliases are 'alt', 'altitude'
            M (dimless): Velocity as Mach number - aliases are 'mach', 'Mach',
                'M_inf', 'mach_number'
            TAS (speed): Velocity as true airspeed - aliases are 'tas',
                'true_airspeed', 'U_inf', 'V_inf'
            CAS (speed): Velocity as calibrated airspeed - aliases are 'cas',
                'calibrated_airspeed'
            EAS (speed): Velocity as equivalent airspeed - aliases are 'eas',
                'equivalent_airspeed'
            L (length): Length scale - aliases are 'ell', 'bylen', 'length',
                'length_scale', 'l'
            Re (dimless): Reynolds number - alternative to velocity or length
                scale but not both - aliases are 'Re_L', 'reynolds_number'
            units (str): Set to 'US' for US units or 'SI' for SI
        """

        # Preprocess needed altitude-based quantities
        # Automatically process altitude through Atmosphere class
        self.byalt = Atmosphere(h=h, units=units, **kwargs)

        # Check for hidden aliases
        M_aliases = ['mach', 'Mach', 'M_inf', 'mach_number']
        if M is None:
            M = __class__._arg_from_alias(M_aliases, kwargs)
        TAS_aliases = ['tas', 'true_airspeed', 'U_inf', 'V_inf', 'VTAS',
                       'vtas']
        if TAS is None:
            TAS = __class__._arg_from_alias(TAS_aliases, kwargs)
        CAS_aliases = ['cas', 'calibrated_airspeed', 'VCAS', 'vcas']
        if CAS is None:
            CAS = __class__._arg_from_alias(CAS_aliases, kwargs)
        EAS_aliases = ['eas', 'equivalent_airspeed', 'VEAS', 'veas']
        if EAS is None:
            EAS = __class__._arg_from_alias(EAS_aliases, kwargs)
        L_aliases = ['ell', 'len', 'length', 'length_scale', 'l']
        if L is None:
            L = __class__._arg_from_alias(L_aliases, kwargs)
        Re_aliases = ['Re_L', 'reynolds_number']
        if Re is None:
            Re = __class__._arg_from_alias(Re_aliases, kwargs)

        # Check if KTAS, KCAS, or KEAS input and append knots unit if so
        KTAS_aliases = ['KTAS', 'ktas', 'knots_true_airspeed']
        if TAS is None:
            KTAS = __class__._arg_from_alias(KTAS_aliases, kwargs)
            TAS = None if KTAS is None else KTAS * unit('knots')
        KCAS_aliases = ['KCAS', 'kcas', 'knots_calibrated_airspeed']
        if CAS is None:
            KCAS = __class__._arg_from_alias(KCAS_aliases, kwargs)
            CAS = None if KCAS is None else KCAS * unit('knots')
        KEAS_aliases = ['KEAS', 'keas', 'knots_equivalent_airspeed']
        if EAS is None:
            KEAS = __class__._arg_from_alias(KEAS_aliases, kwargs)
            EAS = None if KEAS is None else KEAS * unit('knots')

        # Compute velocity-based quantities
        self.byvel = Velocity(self.byalt)

        if M is not None:
            self.byvel.M = M
        elif TAS is not None:
            self.byvel.TAS = TAS
        elif CAS is not None:
            self.byvel.CAS = CAS
        elif EAS is not None:
            self.byvel.EAS = EAS
        elif Re is not None:
            # Velocity is set based on Re and L - set dummy speed for now
            self.byvel.M = 0
        else:
            # Use Mach=0 if no velocity is input
            self.byvel.M = 0*dimless

        # Check that computations are within valid Mach number limits
        M_ = np.atleast_1d(self.byvel.M)
        self._mach_min = 0 * dimless
        self._mach_max = 30 * dimless
        if (M_ < self._mach_min).any() or (self._mach_max < M_).any():
            raise ValueError(
                f"Mach number is out of bounds "
                f"({self._mach_min:.5g} < M_ < {self._mach_max:.5g})"
            )

        # Compute length-scale-based quantities
        self.bylen = Length(self.byalt, self.byvel)

        # If length scale is not input, default to unity with dimentionals unit
        # based on US or SI determination
        if L is None:
            if Re is None:
                L_unit = unit('ft') if self.units == 'US' else unit('m')
                self.bylen.L = 1.0 * L_unit
            else:
                Re *= dimless
                check_dimensionless(Re)
                self.bylen.L = NonDimensional.reynolds_number_length(
                    Re, self.byvel.TAS, self.byalt.nu)
        else:  # length scale is input by user
            self.bylen.L = L

            # If altitude is 0, but length scale is input, determine if US
            # units based on the input length scale
            if units == "":
                if self.byalt.h.magnitude.all() == 0:
                    if check_US_length_units(L):
                        self.units = 'US'

        # Velocity is set based on Re and L
        if Re is not None:
            if (M is not None
                    or TAS is not None
                    or CAS is not None
                    or EAS is not None):
                msg = "Overriding velocity based on Reynolds number."
                warnings.warn(msg)
            self.bylen.Re = Re  # velocity is set based on Re and L

        # Initialize access by full quantity name through .byname.<name>
        self.byvel.byname = AliasAttributes(
            varsobj_arr=[self.byvel, ], varnames_dict_arr=[Velocity.varnames, ]
        )
        self.bylen.byname = AliasAttributes(
            varsobj_arr=[self.bylen, ], varnames_dict_arr=[Length.varnames, ]
        )

        # Set references at base level too
        varsobj_arr = [self.byalt, self.byvel, self.bylen, ]
        varnames_dict_arr = [Atmosphere.varnames,
                             Velocity.varnames,
                             Length.varnames, ]
        own_dict_arr = [
            {key: key for key in dict_.keys()} for dict_ in varnames_dict_arr
        ]

        self.byvar = AliasAttributes(
            varsobj_arr=varsobj_arr,
            varnames_dict_arr=own_dict_arr
        )

        self.varnames = {}
        self.varnames.update(Atmosphere.varnames)
        self.varnames.update(Velocity.varnames)
        self.varnames.update(Length.varnames)

        self.byname = AliasAttributes(
            varsobj_arr=varsobj_arr,
            varnames_dict_arr=varnames_dict_arr
        )

    def tostring(self, full_output=True, units=None, pretty_print=True):
        """String representation of data structure.

        Args:
            full_output (bool): Set to True for full output
            units (str): Set to 'US' for US units or 'SI' for SI
            pretty_print (bool): Pretty print output

        Returns:
            str: String representation
        """
        if units is not None:
            self.units = units

        max_var_chars = max([
            max([len(v) for v in Atmosphere.varnames.values()]),
            max([len(v) for v in Velocity.varnames.values()]),
            max([len(v) for v in Length.varnames.values()]),
        ])
        alti_str = self.byalt.tostring(full_output, self.units,
                                       pretty_print=pretty_print,
                                       max_var_chars=max_var_chars)
        spee_str = self.byvel.tostring(full_output, self.units,
                                       max_var_chars=max_var_chars,
                                       pretty_print=pretty_print)
        leng_str = self.bylen.tostring(full_output, self.units,
                                       max_var_chars=max_var_chars,
                                       pretty_print=pretty_print)

        unit_str = self.units
        ext_str = "full_output=True" if full_output else "full_output=False"
        top_hdr = f"   Flight Condition (units={unit_str}, {ext_str})"
        lin_str = "==========================================================="
        alt_hdr = "------------------  Altitude Quantities  ------------------"
        vel_hdr = "------------------  Velocity Quantities  ------------------"
        len_hdr = "------------------   Length Quantities   ------------------"

        repr_str = (f"{lin_str}\n{top_hdr}\n{lin_str}"
                    f"\n{alt_hdr}" f"\n{alti_str}"
                    f"\n{vel_hdr}" f"\n{spee_str}"
                    f"\n{len_hdr}" f"\n{leng_str}"
                    )

        return repr_str

    @property
    def units(self):
        """Unit system to use: 'SI', 'US', etc.  Available unit systems given
        by dir(unit.sys).

        Returns:
            str: Unit system
        """
        return self.byalt.units

    @units.setter
    def units(self, units):
        """Unit system to use: 'SI', 'US', etc.  Available unit systems given
        by dir(unit.sys).

        Args:
            units (str): Unit system
        """
        self.byalt.units = units

    def redim_force(self, S_ref):
        """Factor to redimensionalize force from force coefficient e.g.,

            .. math::

                L & = C_L q_\\infty S_ref
                  & = C_L \\text{redim_force}

        Args:
            S_ref (area): Reference area

        Returns:
            force: Redimensionalization factor
        """
        check_area_dimensioned(S_ref)
        redim_force = self.byvel.q_inf * S_ref

        # Set to force unit since to_base_units() gives mass*length/time^2
        if self.units == 'US':
            redim_force.ito('lbf')
        elif self.units == 'SI':
            redim_force.ito('N')

        return redim_force

    def redim_moment(self, S_ref, L):
        """Factor to redimensionalize moment from moment coefficient e.g.,

            .. math::

                M & = C_M q_\\infty S_ref L
                  & = C_M \\text{redim_moment}

        Args:
            S_ref (area): Reference area
            L (length): Reference length

        Returns:
            moment: Redimensionalization factor
        """
        check_area_dimensioned(S_ref)
        check_length_dimensioned(L)
        redim_moment = self.byvel.q_inf * S_ref * L

        # Set to force unit since to_base_units() gives mass*length/time^2
        if self.units == 'US':
            redim_moment.ito('ft lbf')
        elif self.units == 'SI':
            redim_moment.ito('m N')

        return redim_moment



"""
Aerospace utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""
import importlib.metadata

from .atmosphere import Atmosphere
from .flightcondition import FlightCondition
from .isentropicflow import IsentropicFlow
from .normalshock import NormalShock
from .units import unit, dimless, printv

__version__ = importlib.metadata.version('flightcondition')



#!/usr/bin/env python
"""Compute properties related to boundary layer flow.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

import numpy as np

from flightcondition.constants import PhysicalConstants as Phys
from flightcondition.isentropicflow import IsentropicFlow


class BoundaryLayer:
    """Compute properties related to boundary layer flow. """

    @staticmethod
    def recovery_temperature_laminar(M, T, y=Phys.gamma_air, Pr=Phys.Pr_air):
        """Adiabiatic wall temperature on infinite flat plate in laminar flow.

        Assumes:
            Adiabatic

        Args:
            M (pressure): Mach number
            T (temperature): Static (ambient) temperature
            y (dimless): ratio of specific heats
            Pr (dimless): Prandtl number

        Returns:
            temperature: Stagnation temperature
        """
        r = Pr**(1/2)
        T0 = IsentropicFlow.T0_by_T(M, y, r)*T
        return T0

    @staticmethod
    def recovery_temperature_turbulent(M, T, y=Phys.gamma_air, Pr=Phys.Pr_air):
        """Adiabiatic wall temperature on infinite flat plate in turbulent
        flow.

        Assumes:
            Adiabatic

        Args:
            M (pressure): Mach number
            T (temperature): Static (ambient) temperature
            y (dimless): ratio of specific heats
            Pr (dimless): Prandtl number

        Returns:
            temperature: Stagnation temperature
        """
        r = Pr**(1/3)
        T0 = IsentropicFlow.T0_by_T(M, y, r)*T
        return T0

    @staticmethod
    def flat_plate_boundary_layer_lamr(x, Re_x):
        """Compute laminar boundary layer for laminar flow over a flat plate.

        Args:
            x (length): Distance along flat plate
            R_x (dimless): Reynolds number at distance along flat plate

        Returns:
            length: boundary layer thickness
        """
        delta_lamr = 5.0*x/Re_x**0.5
        return delta_lamr

    @staticmethod
    def flat_plate_boundary_layer_turb(x, Re_x):
        """Compute turbulent boundary layer for laminar flow over a flat plate.

        Args:
            x (length): Distance along flat plate
            R_x (dimless): Reynolds number at distance along flat plate

        Returns:
            length: boundary layer thickness
        """
        # Derivation of Prandtl (1927),
        # Eq. 7.43 in Boundary Layer Analysis 2nd. Ed. by Schetz and Bowersox
        delta_turb = 0.375*x/Re_x**0.2
        return delta_turb

    @staticmethod
    def flat_plate_skin_friction_coeff_lamr(Re_x):
        """Compute integrated skin friction coefficient for laminar flow over a
        flat plate.  Derived from Blasius solution.

        :math:`C_f(x) = \\tau_w / 0.5 \\rho U^2`

        :math:`\\tau_w` is the wall shear stress, :math:`\\rho` is density,
        and :math:`U_\\infty^2` is freestream velocity.

        Args:
            R_x (dimless): Reynolds number at distance along flat plate

        Returns:
            length: skin friction coefficient
        """
        # Total, integrated, Cf is 2 times local Cf
        Cf_lamr = 2*0.664*Re_x**(-0.5)
        return Cf_lamr

    @staticmethod
    def flat_plate_skin_friction_coeff_turb(Re_x, source='granville1977'):
        """Compute integrated skin friction coefficient for turbulent flow over
        a flat plate.

        :math:`C_f(x) = \\tau_w / 0.5 \\rho U_\\infty^2`

        :math:`\\tau_w` is the wall shear stress, :math:`\\rho` is density,
        and :math:`U_\\infty^2` is freestream velocity.

        Args:
            R_x (dimless): Reynolds number at distance along flat plate

        Returns:
            length: skin friction coefficient
        """
        # See https://www.cfd-online.com/Wiki/Skin_friction_coefficient
        if source == 'granville1977':
            Cf_turb = 0.0776*(np.log10(Re_x) - 1.88)**(-2) + 60*Re_x**(-1)
        elif source == 'ittc1957':
            Cf_turb = 0.075*(np.log10(Re_x) - 2)**(-2)
        elif source == 'schultz_grunov1940':
            Cf_turb = 0.427*(np.log10(Re_x) - 0.407)**(-2.64)
        elif source == 'prandtl_schlichting1932':
            Cf_turb = 0.455*(np.log10(Re_x))**(-2.58)
        elif source == 'prandtl1927':
            Cf_turb = 0.074*Re_x**(-0.2)
        elif source == 'schlichting':  # only valid for Re_x < 1e9
            Cf_turb = (2*np.log10(Re_x) - 0.65)**(-2.3)
        elif source == 'powerlaw':  # only valid for 5e5 < Re_x < 1e7
            Cf_turb = 0.0592*Re_x**(-0.2)
        else:
            raise ValueError(
                f"Input reference is invalid: {source}\n\tSee source code."
            )
        return Cf_turb



#!/usr/bin/env python
"""Compute normal shock quantities.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

# flake8: noqa E201

import numpy as np

from flightcondition.constants import PhysicalConstants as Phys


class NormalShock:
    """Compute normal shock quantities. """

    @staticmethod
    def M2(M1, y=Phys.gamma_air):
        """Compute Mach number downstream of the shock.

        Args:
            M (float): Mach number upstream of shock

        Returns:
            float: Mach number downstream of shock
        """
        M2 = np.sqrt( (M1**2*(y-1) + 2) / (2*y*M1**2 - (y-1)) )
        return M2

    @staticmethod
    def p2_by_p1(M1, y=Phys.gamma_air):
        """Compute ratio of static pressure over shock, downstream/upstream.

        Args:
            M (float): Mach number upstream of shock

        Returns:
            float: ratio of static pressure over shock, downstream/upstream
        """
        p2_by_p1 = (2*y*M1**2)/(y+1) - (y-1)/(y+1)
        return p2_by_p1

    @staticmethod
    def rho2_by_rho1(M1, y=Phys.gamma_air):
        """Compute ratio of density over shock, downstream/upstream.

        Args:
            M (float): Mach number upstream of shock

        Returns:
            float: ratio of density over shock, downstream/upstream
        """
        rho2_by_rho1 = ((y+1)*M1**2) / ((y-1)*M1**2 + 2)
        return rho2_by_rho1

    @staticmethod
    def T2_by_T1(M1, y=Phys.gamma_air):
        """Compute ratio of static pressure over shock, downstream/upstream.

        Args:
            M (float): Mach number upstream of shock

        Returns:
            float: ratio of temperature over shock, downstream/upstream
        """
        T2_by_T1 = (
            (1 + ((y-1)/2)*M1**2)
            *
            (((2*y)/(y-1))*M1**2 - 1)
            /
            (M1**2*((2*y)/(y-1) + (y-1)/2))
        )
        return T2_by_T1

    @staticmethod
    def p02_by_p01(M1, y=Phys.gamma_air):
        """Compute ratio of stagnation pressure over shock, downstream/upstream.

        Args:
            M (float): Mach number upstream of shock

        Returns:
            float: ratio of temperature over shock, downstream/upstream
        """
        p02_by_p01 = (
            (
                (((y+1)/2)*M1**2)
                /
                (1 + ((y-1)/2)*M1**2)
            )**(y/(y-1))
            *
            (
                1
                /
                (((2*y)/(y+1))*M1**2 - (y-1)/(y+1))
            )**(1/(y-1))
        )
        return p02_by_p01



#!/usr/bin/env python
"""Compute non-dimensional flow quantities.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

# flake8: noqa E201

import numpy as np

from flightcondition.constants import PhysicalConstants as Phys
from flightcondition.units import dimless, to_base_units_wrapper

class NonDimensional:
    """Compute non-dimensional flow quantities. """

    @staticmethod
    def knudsen_number(ell, MFP):
        """Compute the Knudsen number :math:`K_n`

        Args:
            ell (length): Length scale
            MFP (length): Mean free path

        Returns:
            dimless: Knudsen number
        """
        Kn = MFP/ell
        return Kn

    @staticmethod
    def mach_number(U, a):
        """Compute Mach number

        Args:
            U (speed): Velocity
            a (speed): Sound speed

        Returns:
            dimless: Mach number
        """
        return U/a

    @staticmethod
    def mach_velocity(M, a):
        """Compute velocity from Mach number

        Args:
            M (dimless): Mach number
            a (speed): Sound speed

        Returns:
            speed: Velocity
        """
        U = M*a if M is not None else None
        return U

    @staticmethod
    def reynolds_number(U, L, nu):
        """Compute Reynolds number :math:`Re`, the ratio of inertial to viscou
        forces.

        Args:
            U (speed): Velocity
            L (length): Length scale
            nu (length^2/time): Kinematic viscosity

        Returns:
            dimless: Reynolds number
        """
        Re_L = U*L/nu
        return Re_L

    @staticmethod
    def reynolds_number_velocity(Re_L, L, nu):
        """Compute velocity from Reynolds number, length, and kinematic
        viscosity.

        Args:
            Re_L (dimless): Reynolds number
            L (length): Length scale
            nu (length^2/time): Kinematic viscosity

        Returns:
            speed: Velocity
        """
        U = Re_L*nu/L
        return U

    @staticmethod
    def reynolds_number_length(Re_L, U, nu):
        """Compute length from Reynolds number, velocity, and kinematic
        viscosity.

        Args:
            Re_L (dimless): Reynolds number
            U (speed): Velocity
            nu (length^2/time): Kinematic viscosity

        Returns:
            speed: Velocity
        """
        L = Re_L*nu/U
        return L

    @staticmethod
    def reynolds_per_length(U, nu, length_unit='in'):
        """Compute Reynolds number divided by length unit.

        Reynolds number is,
            Re = U * L / nu

        For some fluid dynamics solvers the user must input the Reynolds number
        in terms of Reynolds number per-unit-length,
            Re_by_L = Re_L / L_in_grid_units
                      = U * (L/L_in_grid_units) / nu

        Assume the length scale is 5 feet.  When grid units are the same as
        standard length units, such as feet vs. feet, the term
        (L_in_grid_units/L) is unity, and does not change the Reynolds
        number magnitude.  For example,
            L/L_in_grid_units = (5 ft)/(5 ft) = 1 ft/ft

        When grid units differ from standard length units, such as inches vs.
        feet, the term (L_in_grid_units/L) becomes,
            L/L_in_grid_units = (5 ft)/(60 in) = 1/12 ft/in
        which is the conversion factor between feet and inches.

        So, the Re_by_L term becomes
            Re_by_L = Re_L * (1/12 1/in)

        In this case, it is helpful to compute Reynolds number per-unit-length
        in inches (length_unit='in').

        Args:
            U (speed): Velocity
            nu (speed^2/length): Kinematic viscosity
            length_unit (length): Desired length unit as string, ('in', 'mm',
                'cm')

        Returns:
            dimless: Reynolds number
        """
        Re_by_length_unit = U/nu
        if length_unit is not None:
            Re_by_length_unit.ito(f"1/{length_unit}")
        return Re_by_length_unit



"""Command Line Interface (CLI) main function for easy use from the terminal.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

import argparse
import sys

from .flightcondition import FlightCondition
from .units import unit

RUNCMD = 'flightcondition'
SHORT_DESCRIPTION = ("Compute airspeed (true/calibrated/equivalent/Mach), "
                     "atmospheric data, and other flight condition quantities,"
                     " with easy unit conversion.")


def _parse_args(args_arr):
    """Parse arguments.

    Args:
        args (Sequence[str]): argument array

    Returns:
        argparse.Namespace: parsed arguments array
    """
    # Initialize parser and define arguments
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, prog=RUNCMD,
        description=SHORT_DESCRIPTION)
    parser.add_argument(
        '--h', dest='altitude', metavar='', nargs=2, type=str,
        default=[0, 'kft'],
        help=("altitude and length unit, default='0 kft'; aliases include: "
              "--altitude --Altitude"))
    parser.add_argument(
        '--altitude', dest='altitude', metavar='', nargs=2,
        type=str, default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--Altitude', dest='altitude', metavar='', nargs=2,
        type=str, default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--M', dest='M', metavar='', nargs=None, type=float, default=None,
        help=("Mach number, e.g. '0.5'; aliases include: "
              "--mach-number --Mach"))
    parser.add_argument(
        '--mach-number', dest='M', metavar='', nargs=None, type=float,
        default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--Mach', dest='M', metavar='', nargs=None, type=float,
        default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--TAS', dest='TAS', metavar='', nargs=2, type=str, default=None,
        help="true airspeed and speed unit, e.g. '150 knots'")
    parser.add_argument(
        '--CAS', dest='CAS', metavar='', nargs=2, type=str, default=None,
        help="calibrated airspeed and speed unit, e.g. '150 knots'")
    parser.add_argument(
        '--EAS', dest='EAS', metavar='', nargs=2, type=str, default=None,
        help="equivalent airspeed and speed unit, e.g. '150 knots'")
    parser.add_argument(
        '--KTAS', dest='KTAS', metavar='', nargs=None, type=float,
        default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--KCAS', dest='KCAS', metavar='', nargs=None, type=float,
        default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--KEAS', dest='KEAS', metavar='', nargs=None, type=float,
        default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--L', dest='length_scale', metavar='', nargs=2,
        type=str, default=None,
        help=("length scale, e.g. '10 ft'; aliases include: "
              "--length-scale --Length --ell"))
    parser.add_argument(
        '--length-scale', dest='length_scale', metavar='', nargs=2,
        type=str, default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--Length', dest='length_scale', metavar='', nargs=2,
        type=str, default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--ell', dest='length_scale', metavar='', nargs=2,
        type=str, default=None, help=argparse.SUPPRESS)
    parser.add_argument(
        '--units', dest='units', metavar='', nargs=None,
        type=str, default="", help=f"Unit system, i.e. {dir(unit.sys)}")
    parser.add_argument(
        '--no-full-output', dest='no_full_output', default=False,
        action='store_true', help="display abbreviated output")
    parser.add_argument(  # hidden option to turn off pretty print
        '--no-pretty-print', dest='no_pretty_print', default=False,
        action='store_true', help=argparse.SUPPRESS)

    args = parser.parse_args(args_arr)

    return args


def _dimension(num_unit_arr):
    """Parse length 2 array into dimensional unit quantity.

    Args:
        array: float value and unit string

    Returns:
        dimensional: dimensionalized quantity
    """
    val, unit_ = num_unit_arr
    return float(val) * unit(unit_)


def main():
    """Main function. """
    args_arr = sys.argv[1:] if len(sys.argv) > 1 else None
    args = _parse_args(args_arr)
    h_ = _dimension(args.altitude)
    M_ = None if args.M is None else args.M
    TAS_ = None if args.TAS is None else _dimension(args.TAS)
    CAS_ = None if args.CAS is None else _dimension(args.CAS)
    EAS_ = None if args.EAS is None else _dimension(args.EAS)
    TAS_ = TAS_ if args.KTAS is None else args.KTAS*unit('knots')
    CAS_ = CAS_ if args.KCAS is None else args.KCAS*unit('knots')
    EAS_ = EAS_ if args.KEAS is None else args.KEAS*unit('knots')
    L_ = None if args.length_scale is None else _dimension(args.length_scale)
    fc = FlightCondition(h=h_, M=M_, TAS=TAS_, CAS=CAS_, EAS=EAS_, L=L_,
                         units=args.units)

    # Print output but catch common unicode exception
    full_output = not args.no_full_output
    try:
        print(fc.tostring(full_output=full_output,
              pretty_print=(not args.no_pretty_print)))
    except UnicodeEncodeError:
        print(fc.tostring(full_output=full_output, pretty_print=False))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Compute isentropic flow relations.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

# flake8: noqa E201

import numpy as np

from flightcondition.constants import PhysicalConstants as Phys


class IsentropicFlow:
    """Compute isentropic flow relations. """

    # Critical values
    a_star_by_a0 = 0.913
    p_star_by_p0 = 0.528
    rho_star_by_rho0 = 0.634

    @staticmethod
    def p0_by_p(M, y=Phys.gamma_air):
        """Compute ratio of stagnation pressure (brought to rest
        isentropically) to static pressure.

        Args:
            M (float): Mach number
            y (float): ratio of specific heats

        Returns:
            float: ratio of stagnation pressure to static pressure
        """
        T0_by_T = __class__.T0_by_T(M, y)
        p0_by_p = T0_by_T**(y/(y-1))
        return p0_by_p

    @staticmethod
    def T0_by_T(M, y=Phys.gamma_air, r=1):
        """Compute ratio of stagnation temperature (brought to rest
        isentropically) to static temperature.

        Args:
            M (float): Mach number
            y (float): ratio of specific heats
            r (float): recovery factor (if computing recovery temperature)

        Returns:
            float: ratio of stagnation temperature to static temperature
        """
        T0_by_T = 1 + r*((y-1)/2)*M**2
        return T0_by_T

    @staticmethod
    def a0_by_a(M, y=Phys.gamma_air):
        """Compute ratio of stagnation sound speed (brought to rest
        isentropically) to ambient sound speed.

        Args:
            M (float): Mach number

        Returns:
            float: ratio of stagnation sound speed to static sound speed
        """
        a0_by_a = np.sqrt(__class__.T0_by_T(M, y))
        return a0_by_a

    @staticmethod
    def rho0_by_rho(M, y=Phys.gamma_air):
        """Compute ratio of stagnation density (brought to rest
        isentropically) to ambient density.

        Args:
            M (float): Mach number
            y (float): ratio of specific heats

        Returns:
            float: ratio of stagnation density to static density
        """
        T0_by_T = __class__.T0_by_T(M, y)
        rho0_by_rho = T0_by_T**(1/(y-1))
        return rho0_by_rho

    @staticmethod
    def A_star_by_A(M, y=Phys.gamma_air):
        """Compute ratio of area to area at which Mach equals 1.

        Args:
            M (float): Mach number

        Returns:
            float: ratio of area to area at which Mach equals 1
        """
        T0_by_T = __class__.T0_by_T(M, y)
        A_star_by_A = (
            (1/M**2) * ( (2/(y+1))*T0_by_T )**((y+1)/(y-1))
        )**(-1/2)
        return A_star_by_A

    @staticmethod
    def M_from_p0_by_p(p0_by_p, y=Phys.gamma_air):
        """Compute Mach number from ratio of stagnation pressure (brought to
        rest isentropically) to static pressure.

        Args:
            p0_by_p (float): ratio of stagnation pressure to pressure
            y (float): ratio of specific heats

        Returns:
            float: Mach number
        """
        T0_by_T = p0_by_p**((y-1)/y)
        M = __class__.M_from_T0_by_T(T0_by_T)
        return M

    @staticmethod
    def M_from_T0_by_T(T0_by_T, y=Phys.gamma_air):
        """Compute Mach number from ratio of stagnation temperature (brought to
        rest isentropically) to static temperature.

        Args:
            T0_by_T (float): ratio of stagnation temperature to temperature
            y (float): ratio of specific heats

        Returns:
            float: Mach number
        """
        # Derive from: T0_by_T = 1 + ((y-1)/2)*M**2
        M = np.sqrt( (T0_by_T - 1)/((y-1)/2) )
        return M



#!/usr/bin/env python
"""Common classes and functions.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from functools import wraps

import numpy as np

from flightcondition.units import to_base_units_wrapper


class AliasAttributes():
    """Nested alias class to reference quantities by prescribed aliases. """

    def __init__(self, varsobj_arr, varnames_dict_arr):
        """Populate full names and link to variable
        Args:
            varsobj (list): List of objects that holds all of the variables
            varnames_dict (list): List of dictionaries that maps variables to
                their alias names
        """
        # Avoid infinite loop when setting properties
        super().__setattr__("varsobj_arr", varsobj_arr)
        super().__setattr__("varnames_dict_arr", varnames_dict_arr)

    def __dir__(self):
        """Add tab completion for alias names. """
        # Avoid infinite loop when loading varsobj and varnames_dict
        varnames_dict_arr = super().__getattribute__("varnames_dict_arr")
        varnames_arr = []
        for varnames_dict in varnames_dict_arr:
            for varname in varnames_dict.values():
                varnames_arr.append(varname)
        return varnames_arr

    def __getattribute__(self, attr):
        """Get referenced quantity

        Args:
            attr (str): attribute name
        """
        # Avoid infinite loop when loading varsobj and varnames_dict
        varsobj_arr = super().__getattribute__("varsobj_arr")
        varnames_dict_arr = super().__getattribute__("varnames_dict_arr")
        for varsobj, varnames_dict in zip(varsobj_arr, varnames_dict_arr):
            for var, varname in varnames_dict.items():
                if attr == varname:
                    return getattr(varsobj, var)

    def __setattr__(self, attr, attrval):
        """Set referenced quantity

        Args:
            attr (str): attribute name
            attrval (quantity): attribute value
        """
        # Avoid infinite loop when loading varsobj and varnames_dict
        varsobj_arr = super().__getattribute__("varsobj_arr")
        varnames_dict_arr = super().__getattribute__("varnames_dict_arr")
        for varsobj, varnames_dict in zip(varsobj_arr, varnames_dict_arr):
            for var, varname in varnames_dict.items():
                if attr == varname:
                    return setattr(varsobj, var, attrval)


class DimensionalData:
    """Parent class to hold dimensional data"""

    varnames = {}

    def __eq__(self, otherobj):
        """Check equality.
        Returns:
            dict: Dictionary representation of object
        """
        if otherobj.__class__ is not self.__class__:
            return NotImplemented

        return self.asdict() == otherobj.asdict()

    def __str__(self):
        """Output string when object is printed.

        Returns:
            str: Full string output
        """
        return self.tostring(full_output=True)

    def __repr__(self):
        """Output string representation of class object.

        Returns:
            str: Full string output
        """
        return self.tostring(full_output=False)

    def asdict(self):
        """Return class data as dictionary.
        Returns:
            dict: Class data
        """
        obj_dict = {}
        for var, varname in self.varnames.items():
            obj_dict[var] = getattr(self, var)
        return obj_dict

    def print(self, *args, **kwargs):
        """Print tostring() function to stdout. """
        print(self.tostring(*args, **kwargs))

    def tostring(self, full_output=True):
        """Override this function to output string representation of class
        object

        Args:
            full_output (bool): Set to True for full output

        Returns:
            str: String representation
        """
        return ""

    def _vartostr(self, var, var_str, to_units, max_var_chars=0,
                  fmt_val="10.5g", pretty_print=False):
        """Formatted variable string with variable, full name, and value.
        Shortens array output to fit better on screen.

        Args:
            var (Quantity): Variable
            var_str (str): Variable string
            to_units (str): Units to convert to
            fmt_val (str): String for formatting value
            pretty_print (bool): Pretty print format

        Returns:
            str: formatted string
        """
        pp_ = "~P" if pretty_print else ""
        var_name = self.varnames[var_str]
        var.ito(to_units)
        if np.size(var) > 6:
            mag = var.magnitude
            var_val_str = (f"[{mag[0]:{fmt_val}} {mag[1]:{fmt_val}} "
                           f"{mag[2]:{fmt_val}} ... {mag[-3]:{fmt_val}} "
                           f"{mag[-2]:{fmt_val}} {mag[-1]:{fmt_val}}] "
                           f"{var.units:{pp_}}")
        else:
            var_val_str = f"{var:{fmt_val}{pp_}}"
        var_str = f"{var_name:{max_var_chars}s} {var_str:10s} = {var_val_str}"
        return var_str

    @staticmethod
    def _arg_from_alias(alias_list, kwargs_dict):
        """Determine argument from hidden alias list

        Args:
            alias_list (dict): List of possible argument aliases

        Returns:
            unit.Quantity: argument
        """
        for alias in alias_list:
            if alias in kwargs_dict.keys():
                return kwargs_dict[alias]
        else:  # no break
            return None


def _len1array_to_scalar(func):
    """Decorator to output scalar if array is length 1.

    Args:
        func (object): Function to wrap

    Returns:
        object: Scalar or array
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        if isinstance(output.magnitude, np.ndarray):
            if np.size(output) == 1:
                return output[0]
        return output
    return wrapper


def _property_decorators(func):
    """ Combine multiple decorators for properties of the class.

    Equivalent to:

        @property
        @to_base_units_wrapper
        @_len1array_to_scalar
    """
    return property(to_base_units_wrapper(_len1array_to_scalar(func)))
